(require-library simple-tests low-level-macros)
(import simple-tests low-level-macros)
(import-for-syntax
  (only low-level-macros macro-rules once-only with-gensyms))

(compound-test (low-level-macros)
  (define-test (macros?)
    (check
      (define-macro (foo (qux))
        (with-keywords (bar baz)
          `(case ',qux
             ((bar baz) ',qux)
             (else 'no))))
      (eq? (foo (bar)) 'bar)
      (eq? (foo (baz)) 'baz)
      (eq? (foo (qux)) 'no)
      (define-macro (efreeze xpr)
        (with-rename-prefix %
          `(,%lambda () ,xpr)))
      (= ((efreeze 3)) 3)
      (define-macro (ifreeze xpr)
        `(lambda () ,xpr))
      (= ((ifreeze 5)) 5)
      (define-macro (swap! x y)
        `(let ((tmp ,x)) (set! ,x ,y) (set! ,y tmp)))
      (equal? (let ((x 'x) (y 'y)) (swap! x y) (list x y))
              '(y x))
      (= (letrec-macro (((ifreeze xpr) `(lambda () ,xpr))
                        ((efreeze xpr)
                         (with-rename-prefix %
                           `(,%lambda () ,xpr))))
           ((efreeze ((ifreeze 3)))))
         3)
      (equal? (let-macro (((ifreeze xpr) `(lambda () ,xpr))
                          ((efreeze xpr)
                           (with-rename-prefix %
                             `(,%lambda () ,xpr))))
                (list ((efreeze 3)) ((ifreeze 5))))
              '(3 5))

      "VERBOSE IFS"
      (define-macro (verbose-if test (then . xprs) (else . yprs))
        (with-rename-prefix %
          (with-keywords (then else)
            `(,%if ,test
               (,%begin ,@xprs)
               (,%begin ,@yprs)))))
      (define-syntax vif
        (macro-rules (then else)
          ((_ test (then xpr . xprs))
           `(if ,test
              (begin ,xpr ,@xprs)))
          ((_ test (else xpr . xprs))
           `(if ,(not test)
              (begin ,xpr ,@xprs)))
          ((_ test (then xpr . xprs) (else ypr . yprs))
           `(if ,test
              (begin ,xpr ,@xprs)
              (begin ,ypr ,@yprs)))))
      (define (quux x)
        (verbose-if (odd? x) (then "odd") (else "even")))
      (equal? (quux 3) "odd")
      (equal? (quux 4) "even")
      (define (oux)
        (vif #t (then 'true)))
      (define (pux)
        (vif #f (else 'false)))
      (eq? (oux) 'true)
      (eq? (pux) 'false)
      
      "LOW-LEVEL COND"
      (define-syntax my-cond
        (macro-rules (else =>)
          ((_ (else xpr . xprs))
           `(begin ,xpr ,@xprs))
          ((_ (test => xpr))
           (let ((temp test))
             `(if ,temp (,xpr ,temp))))
          ((_ (test => xpr) . clauses)
           (let ((temp test))
             `(if ,temp
                (,xpr ,temp)
                (my-cond ,@clauses))))
          ((_ (test)) test)
          ((_ (test) . clauses)
           (let ((temp test))
             `(if ,temp
                ,temp
                (my-cond ,@clauses))))
          ((_ (test xpr . xprs))
           `(if ,test (begin ,xpr ,@xprs)))
          ((_ (test xpr . xprs) . clauses)
           `(if ,test
              (begin ,xpr ,@xprs)
              (my-cond ,@clauses)))))
      (eq? (my-cond ((> 3 2) 'greater) ((< 3 2) 'less))
           `greater)
      (eq? (my-cond ((> 3 3) 'greater)
               ((< 3 3) 'less)
               (else 'equal))
           'equal)
      (= (my-cond ((assv 'b `((a 1) (b 2) (C 3))) => cadr)
               (else #f))
         2)
      (not (my-cond ((assv 'x `((a 1) (b 2) (C 3))) => cadr)
               (else #f)))

      "ANAPHORIC MACROS"
      (define-macro (alambda args xpr . xprs)
        (with-inject-prefix %
          `(letrec ((,%self (lambda ,args ,xpr ,@xprs)))
             ,%self)))
      (define ! (alambda (n) (if (zero? n) 1 (* n (self (- n 1))))))
      (= (! 5) 120)
      (define-syntax aif
        (macro-rules it ()
          ((_ test consequent)
           `(let ((,it ,test))
              (if ,it ,consequent)))
          ((_ test consequent alternative)
           `(let ((,it ,test))
              (if ,it ,consequent ,alternative)))))
      (define (mist x) (aif (! x) it))
      (= (mist 5) 120)))
  (macros?)

  (define-test (etc?)
    (check
      (define-syntax-rule (freeze x) (lambda () x))
      (= ((freeze 25)) 25))
      (define-macro (square x)
        (once-only (x)
          `(* ,x ,x)))
      (let ((n 4))
        (= (square (begin (set! n (+ n 1)) n)) 25))

      (define-macro (for (var start end) . body)
        (once-only (start end)
          `(do ((,var ,start (add1 ,var)))
                 ((= ,var ,end))
                 ,@body)))
      (define counter
        (let ((state 0))
          (lambda ()
            (set! state (+ state 1))
            state)))
      (let ((lst '()))
        (for (x 0 (counter)) (set! lst (cons x lst)))
        (equal? lst '(0)))
      (define-macro (times a b)
        (with-gensyms (x y)
          `(let ((,x ,a) (,y ,b))
             (* ,x ,y))))
      (= (times 3 5) 15))
  (etc?)
  )



